﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace Remake.Engine
{
    public interface IParser<T>
    {
        T Parse(TextReader reader);
    }

    public interface IFormatter<T>
    {
        void Format(TextWriter writer, T item);
    }

    public abstract class ListParserBase<T> : IParser<IList<T>>
    {
        public virtual IList<T> Parse(TextReader reader)
        {
            var result = new List<T>();
            for (; ; )
            {
                var item = ReadSingle(reader);
                if (item == null)
                {
                    break;
                }
                result.Add(item);
            }
            return result;
        }

        public abstract T ReadSingle(TextReader reader);
    }

    public class CompetitionParser : ListParserBase<Competition>
    {
        private CompetitionDateFormat dateFormat = new CompetitionDateFormat();
        private CompetitionCostFormat costFormat = new CompetitionCostFormat();

        public override Competition ReadSingle(TextReader reader)
        {
            var name = reader.ReadLine();
            if (name == null)
            {
                return null;
            }
            var date = dateFormat.Parse(reader.ReadLine());
            var cost = costFormat.Parse(reader.ReadLine());
            var contestantsFile = reader.ReadLine();
            return new Competition() {
                Name = name,
                Cost = cost,
                Date = date,
                ContestantFile = contestantsFile
            };
        }
    }

    public class ContestantParser : ListParserBase<Contestant>
    {
        public override Contestant ReadSingle(TextReader reader)
        {
            var name = reader.ReadLine();
            if (name == null)
            {
                return null;
            }
            var birthYear = int.Parse(reader.ReadLine());
            var achievements = reader.ReadLine();
            return new Contestant()
            {
                Name = name,
                BirthYear = birthYear,
                Achievements = achievements
            };
        }
    }

    public class BetParser : ListParserBase<Bet>, IFormatter<IEnumerable<Bet>>
    {
        public override Bet ReadSingle(TextReader reader)
        {
            var line = reader.ReadLine();
            if (line == null)
            {
                return null;
            }
            var parts = line.Split();
            var action = parts[0];
            var competition = parts[1];
            var names = new List<string>();
            for (var i = 3; i < parts.Length; i += 2)
            {
                names.Add(parts[i - 1] + " " + parts[i]);
            }
            return new Bet()
            {
                Action = ParseEnum<BetAction>(action),
                Competition = competition,
                Contestants = names
            };
        }
        public static T ParseEnum<T>(string value)
        {
            return (T)Enum.Parse(typeof(T), value, ignoreCase: true);
        }

        public void Format(TextWriter writer, IEnumerable<Bet> items)
        {
            foreach (var item in items)
            {
                writer.Write(item.Action.ToString());
                writer.Write(" ");
                writer.Write(item.Competition);
                foreach (var contestant in item.Contestants)
                {
                    writer.Write(" ");
                    writer.Write(contestant);
                }
                writer.WriteLine();
            }
        }
    }

    public class BetGroupFormat : IParser<BetGroup>, IFormatter<BetGroup>
    {
        private BetParser betParser = new BetParser();
        private CompetitionDateFormat dateFormat = new CompetitionDateFormat();

        public void Format(TextWriter writer, BetGroup group)
        {
            writer.WriteLine(group.ManagerName);
            writer.WriteLine(dateFormat.Format(group.Date));
            betParser.Format(writer, group.Bets);
        }

        public BetGroup Parse(TextReader reader)
        {
            var manager = reader.ReadLine();
            var date = dateFormat.Parse(reader.ReadLine());
            var bets = betParser.Parse(reader);
            return new BetGroup()
            {
                ManagerName = manager,
                Date = date,
                Bets = bets
            };
        }
    }


    public class CompetitionCostFormat
    {
        private string suffix = " kupon";
        public int Parse(string input)
        {
            return int.Parse(input.Substring(0, input.Length - suffix.Length));
        }
        public string Format(int input)
        {
            return input.ToString() + suffix;
        }
    }
    public class CompetitionDateFormat
    {
        private string format = "yyyy. mm. dd.";
        private string prefix = "2009. ";
        public DateTime Parse(string input)
        {
            return DateTime.ParseExact(prefix + input, format, CultureInfo.InvariantCulture);
        }
        public string Format(DateTime input)
        {
            return input.ToString(format, CultureInfo.InvariantCulture).Substring(prefix.Length);
        }
    }
}
