﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;

namespace Remake.Engine
{
    public class Simulator
    {
        public string ManagerName { get; protected set; }
        public int BaseCredits { get; protected set; }
        public int AllCredits { get; protected set; }

        private Repository repository;
        private IList<BetRecord> records = new List<BetRecord>();
        private IDictionary<Contestant, BetRecord> lastRecord = new Dictionary<Contestant, BetRecord>();
        public IList<Bet> Bets { get; protected set; }

        private PrizeCalculator prizes = new PrizeCalculator();

        public Simulator(Repository repository, string managerName)
        {
            this.repository = repository;
            this.ManagerName = managerName;
            Bets = new List<Bet>();
            BaseCredits = 200;
            AllCredits = 200;
        }

        public void SelectDate(DateTime date)
        {
            var betGroups = repository.ReadBetGroups()
                .Where(g => g.ManagerName == ManagerName).ToList();
            if (betGroups.Any(g => g.Date >= date))
            {
                throw new BetsAlreadyExistAfterThisException();
            }
            var betGroupsToProcess = betGroups
                .Where(g => g.Date < date)
                .OrderBy(g => g.Date);
            foreach (var betGroup in betGroupsToProcess)
            {
                foreach (var bet in betGroup.Bets)
                {
                    HandleBet(bet, betGroup.Date);
                }
            }
            
        }

        public void HandleBet(Bet bet, DateTime date)
        {
            switch (bet.Action)
            {
                case BetAction.fogad:
                    PlaceBet(bet.Competition, bet.Contestants, date);
                    break;
                case BetAction.vissza:
                    CancelBet(bet.Competition, bet.Contestants, date);
                    break;
            }
        }

        public void PlaceBet(string competitionName, IList<string> betContestants, DateTime date)
        {
            var competition = repository.GetCompetitionByName(competitionName);
            if (competition.Date < date)
            {
                throw new BetPeriodOverException(competitionName);
            }
            var competitionContestants = competition.Contestants;

            var contestants = betContestants.Select(competitionContestants.GetContestantByName).ToList();

            foreach (var contestant in contestants)
            {
                if (lastRecord.ContainsKey(contestant))
                {
                    throw new AlreadyBetOnPlayerException(contestant.Name);
                }
            }
            var totalBets = competition.Cost * betContestants.Count;
            if (AllCredits < totalBets)
            {
                throw new NotEnoughCreditsException();
            }
            foreach (var contestant in contestants)
            {
                PlaceBetUnsafe(contestant, competition.Cost);
            }
            Bets.Add(new Bet() {
                Action = BetAction.fogad,
                Competition = competitionName,
                Contestants = betContestants
            });
        }

        public void StoreBets(DateTime date)
        {
            repository.WriteNextBetGroup(new BetGroup() {
                ManagerName = ManagerName,
                Date = date,
                Bets = Bets
            });
        }

        public void CancelBet(string competitionName, IList<string> betContestants, DateTime date)
        {
            var competition = repository.GetCompetitionByName(competitionName);
            if (competition.Date < date)
            {
                throw new BetPeriodOverException(competitionName);
            }
            var competitionContestants = competition.Contestants;

            var contestants = betContestants.Select(competitionContestants.GetContestantByName).ToList();

            foreach (var contestant in contestants)
            {
                if (!lastRecord.ContainsKey(contestant))
                {
                    throw new NoBetOnPlayerException(contestant.Name);
                }
            }

            foreach (var contestant in contestants)
            {
                var betRecord = lastRecord[contestant];
                CancelBetUnsafe(betRecord);
            }

            Bets.Add(new Bet()
            {
                Action = BetAction.vissza,
                Competition = competitionName,
                Contestants = betContestants
            });

        }

        public void CloseDay(Competition competition)
        {
            foreach (var contestant in competition.Contestants.Contestants)
            {
                var place = competition.Places.IndexOf(contestant.Name) + 1;
                if (place > 0) { 
                    BetRecord bet;
                    if (lastRecord.TryGetValue(contestant, out bet))
                    {
                        var prize = prizes.DeterminePrize(bet.Credits, place);
                        lastRecord.Remove(contestant);
                        GiveCredits(prize);
                    }
                }
            }

        }

        private bool TakeCreditsReturnIfWonUsed(int credits)
        {
            AllCredits -= credits;
            if (BaseCredits < credits)
            {
                BaseCredits = 0;
                return true;
            }
            BaseCredits -= credits;
            return false;
        }

        private void GiveCredits(int credits)
        {
            AllCredits += credits;
        }
        private void PlaceBetUnsafe(Contestant contestant, int betSize)
        {
            var usedAnyWonCredits = TakeCreditsReturnIfWonUsed(betSize);

            var record = new BetRecord()
            {
                Contestant = contestant,
                Credits = betSize,
                UsedAnyWonCredits = usedAnyWonCredits,
                Action = BetAction.fogad
            };
            records.Add(record);
            lastRecord[contestant] = record;
        }

        private CancelBetResult CancelBetUnsafe(BetRecord bet)
        {
            lastRecord.Remove(bet.Contestant);
            records.Add(new BetRecord()
            {
                Action = BetAction.vissza,
                Contestant = bet.Contestant,
                Credits = bet.Credits
            });
            if (!bet.UsedAnyWonCredits)
            {
                return CancelBetResult.OkNincsVisszaterites;
            }
            GiveCredits(bet.Credits);
            return CancelBetResult.OkVisszateritve;
        }
    }

    public enum CancelBetResult
    {
        OkNincsVisszaterites,
        OkVisszateritve
    }

    public class BetRecord
    {
        public BetAction Action { get; set; }
        public Contestant Contestant { get; set; }
        public int Credits { get; set; }
        public bool UsedAnyWonCredits { get; set; }
    }

    public class NotEnoughCreditsException : Exception
    {
    }
    public class AlreadyBetOnPlayerException : Exception
    {
        public AlreadyBetOnPlayerException(string message) : base(message)
        {
        }
    }
    public class NoBetOnPlayerException : Exception
    {
        public NoBetOnPlayerException(string message) : base(message)
        {
        }
    }
    public class BetsAlreadyExistAfterThisException : Exception
    {
    }
    public class BetPeriodOverException : Exception
    {
        public BetPeriodOverException(string message) : base(message)
        {
        }
    }
}
