﻿using Remake.Engine;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Remake
{

    class Program
    {
        static void Main(string[] args)
        {
            Directory.SetCurrentDirectory(Path.Combine(Directory.GetCurrentDirectory(), "Tests"));
            new Program().Run();
        }

        private Repository repository;
        private CompetitionDateFormat dateFormat = new CompetitionDateFormat();
        private CompetitionCostFormat costFormat = new CompetitionCostFormat();

        public Program()
        {
            repository = Repository.FromFile("verseny.txt");
        }

        public void Run()
        {
            PromptLoop("Milyen minosegben szeretned hasznalni a programot?", PromptEnum<UserKind>, user =>
            {
                switch (user)
                {
                    case UserKind.menedzser:
                        RunManager();
                        break;
                    case UserKind.munkatars:
                        RunEmployee();
                        break;
                }
            });
            Console.WriteLine("Viszlat!");
        }

        private void RunManager()
        {
            Console.WriteLine("A szabalyok megismeresehez kerem olvassa el a mellekelt tajekoztatot.");

            PromptLoop("Mit szeretnel csinalni?",
                PromptEnum<ManagerAction>, action =>
                {
                    switch (action)
                    {
                        case ManagerAction.versenyszamok:
                            PrintCompetitions();
                            break;
                        case ManagerAction.szam:
                            var competitionName = PromptLine("Melyik versenyszam?");
                            PrintCompetition(competitionName);
                            break;
                        case ManagerAction.fogad:
                            RunManagerBet();
                            break;
                    }
                }
            );
        }

        private void RunManagerBet()
        {
            HandleErrors(() =>
            {
                var managerName = PromptLine("A te manager neved:");
                var dateStr = PromptLine("A nap amelyen fogadnal:");
                var date = dateFormat.Parse(dateStr);
                var simulator = new Simulator(repository, managerName);
                simulator.SelectDate(date);
                PromptLoop("Mit szeretnel tenni?", PromptEnum<BetAction>, action =>
                {
                    HandleErrors(() =>
                    {
                        ReadAndSimulateGivenAction(action, simulator, date);
                    });
                }, () => {
                    Console.WriteLine("Kuponjaid: {0} ebbol kezdeti: {1}", simulator.AllCredits, simulator.BaseCredits);
                });
                simulator.StoreBets(date);
            });
        }

        private void ReadAndSimulateGivenAction(BetAction action, Simulator simulator, DateTime date)
        {
            var competitionName = PromptLine("versenyszam:");
            var competition = repository.GetCompetitionByName(competitionName);
            var contestants = new List<string>();
            UntilBack(() =>
            {
                var contestantName = PromptLine("Add meg a kovetkezo versenyzot");
                HandleErrors(() =>
                {
                    competition.Contestants.GetContestantByName(contestantName);
                });
                contestants.Add(contestantName);
            });
            simulator.HandleBet(new Bet()
            {
                Action = action,
                Competition = competitionName,
                Contestants = contestants
            }, date);
        }

        private void UntilBack(Action action)
        {
            try
            {
                for (; ; )
                {
                    action();
                }
            }
            catch (BackException)
            {
                return;
            }
        }

        private void HandleErrors(Action action)
        {
            try
            {
                action();
            }
            catch (EntityNotFoundException ex)
            {
                Console.WriteLine(string.Format("A {0} {1} nem talalhato", ex.EntityType, ex.Message));
            }
            catch (NotEnoughCreditsException ex)
            {
                Console.WriteLine("Nincs eleg kupon erre a fogadasra");
            }
            catch (AlreadyBetOnPlayerException ex)
            {
                Console.WriteLine(string.Format("Mar fogadtal {0}-ra", ex.Message));
            }
            catch (NoBetOnPlayerException ex)
            {
                Console.WriteLine(string.Format("Meg nem fogadtal {0}-ra", ex.Message));
            }
            catch (BetsAlreadyExistAfterThisException ex)
            {
                Console.WriteLine("Mar fogadtal ezen vagy kesobbi napon");
            }
            catch (BetPeriodOverException ex)
            {
                Console.WriteLine(string.Format("A {0} versenynek mar vege", ex.Message));
            }
        }
        private void PrintCompetition(string competitionName)
        {
            HandleErrors(() =>
            {
                var comp = repository.GetCompetitionByName(competitionName);
                PrintContestants(comp.Contestants.Contestants);
            });
        }

        private void PrintContestants(IList<Contestant> conts)
        {
            PrintPaged(conts, (c, i) => string.Format("{0,-20} {1,-8} {2}", c.Name, c.BirthYear, c.Achievements));
        }

        private void PrintCompetitions()
        {
            Console.WriteLine("Versenyszamok:");
            PrintPaged(repository.Competition, (c, i) =>
                string.Format("{0,2}. {1,-20} {2,-8} {3,10}",
                    i + 1, c.Name, dateFormat.Format(c.Date), costFormat.Format(c.Cost)
                )
            );
        }

        private void PrintPaged<T>(IEnumerable<T> items, Func<T, int, string> format)
        {
            var i = 0;
            foreach (var item in items)
            {
                if (i % 20 == 19)
                {
                    Console.WriteLine("Lapozashoz nyomjon entert...");
                    Console.ReadLine();
                }
                Console.WriteLine(format(item, i));
                i++;
            }
        }

        private void RunEmployee()
        {
            var betGroups = repository.ReadBetGroups();
            var managers = new Dictionary<string, Simulator>();
            var hasLastDate = false;
            DateTime lastDate = default(DateTime);
            var allDates = betGroups.Select(g => g.Date).Concat(repository.Competition.Select(c => c.Date)).Distinct().OrderBy(d => d).ToList();
            foreach (var date in allDates)
            {
                foreach (var betGroup in betGroups.Where(g => g.Date == date))
                {
                    var simulator = GetOrCreateSimulator(managers, betGroup);
                    foreach (var item in betGroup.Bets)
                    {
                        simulator.HandleBet(item, betGroup.Date);
                    }
                }
                foreach (var competition in repository.Competition.Where(c => c.Date == date))
                {
                    foreach (var simulator in managers.Values)
                    {
                        simulator.CloseDay(competition);
                    }
                }
                WriteResultFile(managers, date);
            }
            WriteFinalResultFile(managers);
        }

        private void WriteFinalResultFile(Dictionary<string, Simulator> managers)
        {
            using (var f = File.OpenWrite("vegeredm.txt"))
            using (var writer = new StreamWriter(f))
            {
                foreach (var item in managers.Values.OrderByDescending(x => x.AllCredits))
                {
                    var payouts = Math.Max(item.AllCredits - 200, 0);
                    writer.WriteLine(string.Format("{0} {1}", item.ManagerName, payouts));
                }
            }
        }

        private void WriteResultFile(Dictionary<string, Simulator> managers, DateTime lastDate)
        {
            var datePart = lastDate.ToString("mmdd");
            var filename = string.Format("nap{0}.txt", datePart);
            using (var f = File.OpenWrite(filename))
            using (var writer = new StreamWriter(f))
            {
                foreach (var item in managers.Values.OrderByDescending(x => x.AllCredits))
                {
                    writer.WriteLine(string.Format("{0} {1} {2}", item.ManagerName, item.BaseCredits, item.AllCredits - item.BaseCredits));
                }
            }
        }

        private Simulator GetOrCreateSimulator(Dictionary<string, Simulator> managers, BetGroup betGroup)
        {
            Simulator simulator;
            managers.TryGetValue(betGroup.ManagerName, out simulator);
            if (simulator == null)
            {
                simulator = new Simulator(repository, betGroup.ManagerName);
                managers[betGroup.ManagerName] = simulator;
            }

            return simulator;
        }

        private string PromptLine(string question)
        {
            Console.WriteLine(question);
            Console.Write("> ");
            var answer = Console.ReadLine();
            if (IsAnswerBack(answer))
            {
                throw new BackException();
            }
            return answer;
        }

        private void PromptLoop<T>(string question, Func<string, T> prompt, Action<T> action, Action printInfo = null)
        {
            for (; ; )
            {
                printInfo?.Invoke();
                T result;
                try
                {
                    result = prompt(question);
                }
                catch (BackException)
                {
                    return;
                }
                try
                {
                    action(result);
                }
                catch (BackException)
                {
                    // ignore
                }
            }
        }

        private T PromptEnum<T>(string question) where T : struct
        {
            for (; ; )
            {
                var answers = string.Join(", ", Enum.GetNames(typeof(T)));
                var answer = PromptLine(question + " (" + answers + ")");
                var result = ParseEnumByPrefix<T>(answer);
                if (result.HasValue)
                {
                    return result.Value;
                }
                Console.WriteLine("Nincs ilyen valaszlehetoseg");
            }
        }

        private Option<T> ParseEnumByPrefix<T>(string value)
        {
            return Utils.GetByPrefix(Enum.GetValues(typeof(T)).Cast<T>(), value);
        }

        private static bool IsAnswerBack(string answer)
        {
            return answer == "" || answer == "/";
        }
    }

    public enum ManagerAction
    {
        versenyszamok, szam, fogad
    }

    public enum UserKind
    {
        menedzser, munkatars
    }

    public class BackException : Exception { }
}
