﻿using Remake.Engine;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using static Remake.Program;

namespace Remake
{
    public class ContestantCollection
    {
        public IList<Contestant> Contestants { get; protected set; }
        public ContestantCollection(IList<Contestant> contestants)
        {
            Contestants = contestants;
        }

        public Option<Contestant> GetContestantByNamePrefix(string name)
        {
            return Utils.GetByPrefix(Contestants, name, c => c.Name);
        }

        public Contestant GetContestantByName(string name)
        {
            var result = Contestants.SingleOrDefault(c => c.Name == name);
            if (result == null)
            {
                throw new EntityNotFoundException(EntityType.versenyzo, name);
            }
            return result;
        }
    }

    public class Repository
    {
        public IList<Competition> Competition;

        public static Repository FromFile(string filename)
        {
            var repo = new Repository();
            repo.ReadRepository(filename);
            return repo;
        }

        private void ReadRepository(string filename)
        {
            Competition = ReadFile(filename, new CompetitionParser());
            foreach (var item in Competition)
            {
                item.Contestants = GetContestantsOf(item);
                item.Places = GetPlacesOf(item);
            }
        }

        private IList<string> GetPlacesOf(Competition item)
        {
            var filename = item.ContestantFile.Replace("verseny", "eredm");
            return File.ReadAllLines(filename).Where(n => !string.IsNullOrWhiteSpace(n)).ToList();
        }

        private ContestantCollection GetContestantsOf(Competition competition)
        {
            var contestants = ReadFile(competition.ContestantFile, new ContestantParser());
            return new ContestantCollection(contestants);
        }

        private T ReadFile<T>(string filename, IParser<T> parser)
        {
            using (var f = File.OpenRead(filename))
            {
                return parser.Parse(new StreamReader(f));
            }
        }

        public Competition GetCompetitionByName(string name)
        {
            var result = Competition.SingleOrDefault(c => c.Name == name);
            if (result == null)
            {
                throw new EntityNotFoundException(EntityType.verseny, name);
            }
            return result;
        }

        public IList<BetGroup> ReadBetGroups()
        {
            var result = new List<BetGroup>();
            for (var i = 1; ; i++)
            {
                var filename = BetFilename(i);
                if (!File.Exists(filename))
                {
                    return result;
                }
                var bg = ReadFile(filename, new BetGroupFormat());
                result.Add(bg);
            }
        }

        private string NextBetGroup()
        {
            for (var i = 1; ; i++)
            {
                var filename = BetFilename(i);
                if (!File.Exists(filename))
                {
                    return filename;
                }
            }
        }

        private static string BetFilename(int i)
        {
            return string.Format("fogad{0}.txt", i);
        }

        public void WriteNextBetGroup(BetGroup betGroup)
        {
            var filename = NextBetGroup();
            using (var f = File.OpenWrite(filename))
            using (var writer = new StreamWriter(f))
            {
                new BetGroupFormat().Format(writer, betGroup);
            }
        }
    }

    public class EntityNotFoundException : Exception
    {
        public EntityType EntityType { get; protected set; }
        public EntityNotFoundException(EntityType type, string msg) : base(msg)
        {
            EntityType = type;
        }
    }
    public enum EntityType
    {
        verseny, versenyzo
    }
}
